"""
-----------------------------------------------------------------------------
First coded by Luca - De Whiskey's - De Vitis <luca@debian.org> on 2003.05.11
-----------------------------------------------------------------------------
$Id: ZopeCTL.py,v 1.1.1.1 2003/07/02 13:55:39 luca Exp $
"""
__version__ = '$Revision: 1.1.1.1 $'
__source__ = '$Source: /cvsroot/pkg-zope/zopectl/ZopeCTL.py,v $'
__license__ = """
    Copyright (C) 2003 Luca - De Whiskey's - De Vitis <luca@debian.org>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

__all__ = [
    'ZopeInstanceEvent',
    'ZopeInstance',
    'ZopeControlEvent',
    'ZopeControlDisaster',
    'ZopeControlFail',
    'ZopeControl',
]

import os
import sys
from glob import glob
from ConfigParser import ConfigParser
from signal import SIGTERM, SIGINT, SIGHUP, SIGUSR2
from time import sleep
from pwd import getpwnam
from grp import getgrnam
from getopt import getopt

def _ (string):
    return string

class ZopeInstanceConf:
    """
    REQUIRED KEYS
      basename              The configuration file basename without any file
                            extension.
      section               The name of the section that described the
                            instance.
      name                  The name of the instance. A reasonable default is
                            %(basename)s/%(section)s.
      user                  The user name that will be used to run the
                            instance.
      group                 The group name that will be used to run the
                            instance. A reasonable default is %(user)s
      data-fs-in            The location of the default Data.fs file for new
                            instances.
      instance-home         Where instance files are located. A reasonable
                            default is /var/lib/zope/instance/%(name)s.
      zope-home             Where z2.py, zpasswd.py an the ZServer component
                            are located.
      software-home         Where other components and product are located.
    OPTIONAL KEYS
      access-file           This instance access file location.
      pcgi-error-log        The value of PCGI_ERROR_LOG in the pcgi resource
                            file.
      pcgi-exe              The value of PCGI_EXE in the pcgi resource file.
      pcgi-module-path      The value of PCGI_MODULE_PATH in the pcgi resource
                            file.
      pcgi-name             The value of PCGI_NAME in the pcgi resource file.
      pcgi-pid-file         The value of PCGI_PID_FILE in the pcgi resource
                            file.
      pcgi-publisher        The value of PCGI_PUBLISHER in the pcgi resource
                            file.
      pcgi-resource-file    The PCGI resource file location.
      pcgi-socket-file      The value of PCGI_SOCKET_FILE in the pcgi resource
                            file.
      fastcgi-path-or-port  A UNIX socket file, or a port number that will be
                            used by the FastCGI server.
      z2-detailed-log-file  The instance detailed log file location.
      event-log-file        The instance event log file.
      event-log-format      The instance custom event logs format.
      event-log-severity    The instance minimum accepted level for event logs.
      zsyslog               The syslog device on which send event logs.
      zsyslog-facility      The facility to use sending event logs to syslog.
      zsyslog-server        The syslog UDP address:port to use to send event
      zsyslog-access        Like zsyslog, but for access logs rather then
                            event logs.
      zsyslog-access-facility  Like zsyslog-facility, but for access logs
                               rather then event logs.
      zsyslog-access-server  Like zsyslog-server, but for access logs rather
                             then event logs.
      z2-log-file           The instance log file location.
      z2-pid-file           The instance PID file location.
      profile-publisher     A file name to which zope writes the profiling
                            information of each request.
      suppress-accessrule   If set, all SiteRoots are suppressed (used by
                            SiteAccess products).
      suppress-siteroot     If set, all site access rules are suppressed (used
                            by SiteAccess products).
      zsp-ownerous-skip     If set, will cause the Zope Security Policy to
                            skip checks relating to ownership, for servers on
                            which ownership is not important.
      zsp-authenticated-skip  If set, will cause the Zope Security Policy to
                              skip checks relating to authentication, for
                              servers which serve only anonymous content.
      zsession-add-notify   An optional full Zope path name of a callable
                            object to be set as the "script to call on object
                            addition" of the session_data transient object
                            container created in temp_folder at startup. 
      zsession-del-notify   An optional full Zope path name of a callable
                            object to be set as the "script to call on object
                            deletion" of the session_data transient object
                            container created in temp_folder at startup. 
      zsession-timeout-mins  The number of minutes to be used as the "data
                             object timeout" of the
                             "/temp_folder/session_data" transient object
                             container.
      zsession-object-limit  The number of items to use as a "maximum number
                             of subobjects" value of the "/temp_folder"
                             session data transient object container.
      webdav-source-port-clients  Setting this variable enables the retrieval
                                  of the document source through the standard
                                  HTTP port instead of the WebDAV port. The
                                  value of this variable is a regular
                                  expression that is matched against the
                                  user-agent string of the client. 
      stx-default-level     Set this variable to change the default level for
                            <Hx> elements.
      zope-dtml-request-autoquote  Set this variable to one of 'no', '0' or
                                   'disabled' to disable autoquoting of
                                   implicitly retrieved REQUEST data that
                                   contain a '<' when used in a dtml-var
                                   construction.
      base-port             A value of [address:]port type that will be used
                            to generate HTTP, FTP and monitor [address]:port
                            values. Multiple values separated by white space
                            are allowed.
      http-port             A value of [address:]port type that will be used
                            to listen for incoming HTTP connections. Multiple
                            values separated by white space are allowed.
      webdav-port           A value of [address:]port type that will be used
                            to listen for incoming WebDAV connections.
                            Multiple values separated by white space are
                            allowed.
      ftp-port              A value of [address:]port type that will be used
                            to listen for incoming FTP connections. Multiple
                            values separated by white space are allowed.
      icp-port              A value of [address:]port type that will be used
                            for load balancing between multiple back-end Zope
                            servers and a ICP-aware front end. Multiple values
                            separated by white space are allowed.
      monitor-port          A value of [address:]port type that will be used
                            for the secure monitor server port. Multiple
                            values separated by white space are allowed.
      use-management-process  A value that will be use to decide whether to
                            start the management process or not.
      check-interval        A value that will be used as interpreter check
                            interval (in seconds).
      dns                   A value that will be used as the IP address of the
                            DNS server.
      listen                A value that will be used as the IP address to
                            listen on for incoming connections.
      locale                A value that will be used as the locale identifier
                            for the instance message's localization.
      number-of-tread       A value that will be used as the number of thread
                            to use if ZODB3 is used.
      zope-database-quota   An integer number of bytes.
      zope-read-only        If this variable is set, then the database is
                            opened in read only mode. If this variable is set
                            to a string parsable by DateTime.DateTime, then
                            the database is opened read-only as of the time
                            given.
    OTHER KEYS
      Other neither REQUIRED, nor OPTIONAL keys may be provided as well.
      Those key will be ignored for the purpose of zopectl or instances
      configuration, but may provide useful values in format strings.
    """
    def __init__(self, *config):
        self.__headers = ['command', 'flag', 'type', 'value']
        self.__true = ('1', 'yes', 'true', 'on')
        self.__false = ('0', 'no', 'false', 'off')
        self.__table = {
            #'access-encoding':              ['pwd', '-e',   None,   None],
            'access-file':                  ['pwd', None,   None,   None],
            #'access-restriction':           ['pwd', '-d',   None,   None],
            #'access-username':              ['pwd', '-u',   None,   None],
            'basename':                     [None,  None,   None,   None],
            'check-interval':               ['z2',  '-i',   None,   None],
            'debug':                        ['z2',  '-D',   'flag', None],
            'dns':                          ['z2',  '-d',   None,   None],
            'data-fs-in':                   [None,  None,   None,   None],
            'force-http-connection-close':  ['z2',  '-C',   'flag', None],
            'group':                        [None,  None,   None,   None],
            'instance-home':                ['z2',  None,   None,   None],
            'software-home':                ['z2',  None,   None,   None],
            'zope-home':                    ['z2',  '-z',   None,   None],
            'listen':                       ['z2',  '-a',   None,   None],
            'locale':                       ['z2',  '-L',   None,   None],
            'name':                         [None,  None,   None,   None],
            'number-of-tread':              ['z2',  '-t',   None,   None],
            'pcgi-error-log':               ['pcgi',None,   None,   None],
            'pcgi-exe':                     ['pcgi',None,   None,   None],
            'pcgi-module-path':             ['pcgi',None,   None,   None],
            'pcgi-name':                    ['pcgi',None,   None,   None],
            'pcgi-pid-file':                ['pcgi',None,   None,   None],
            'pcgi-publisher':               ['pcgi',None,   None,   None],
            'pcgi-resource-file':           ['z2',  '-p',   None,   None],
            'pcgi-socket-file':             ['pcgi',None,   None,   None],
            'pcgi-wrapper':                 [None,  None,   None,   None],
            'profile-publisher':            ['z2',  None,   None,   None],
            'suppress-accessrule':          ['z2',  None,   None,   None],
            'suppress-siteroot':            ['z2',  None,   None,   None],
            'zsp-ownerous-skip':            ['z2',  None,   None,   None],
            'zsp-authenticated-skip':       ['z2',  None,   None,   None],
            'zsession-add-notify':          ['z2',  None,   None,   None],
            'zsession-del-notify':          ['z2',  None,   None,   None],
            'zsession-timeout-mins':        ['z2',  None,   None,   None],
            'zsession-object-limit':        ['z2',  None,   None,   None],
            'webdav-source-port-clients':   ['z2',  None,   None,   None],
            'stx-default-level':            ['z2',  None,   None,   None],
            'zope-dtml-request-autoquote':  ['z2',  None,   None,   None],
            'zope-database-quota':          ['z2',  None,   None,   None],
            'zope-read-only':               ['z2',  None,   None,   None],
            'base-port':                    ['z2',  '-P',   'multi',None],
            'ftp-port':                     ['z2',  '-f',   'multi',None],
            'http-port':                    ['z2',  '-w',   'multi',None],
            'icp-port':                     ['z2',  '--icp','multi',None],
            'monitor-port':                 ['z2',  '-m',   'multi',None],
            'webdav-port':                  ['z2',  '-W',   'multi',None],
            'fastcgi-path-or-port':         ['z2',  '-F',   'multi',None],
            'read-only':                    ['z2',  '-r',   'flag', None],
            'section':                      [None,  None,   None,   None],
            'use-management-process':       ['z2',  '-Z',   'bool', None],
            'user':                         ['z2',  '-u',   None,   None],
            'z2-detailed-log-file':         ['z2',  '-M',   None,   None],
            'event-log-format':             ['z2',  None,   None,   None],
            'event-log-file':               ['z2',  None,   None,   None],
            'event-log-severity':           ['z2',  None,   None,   None],
            'zsyslog':                      ['z2',  None,   None,   None],
            'zsyslog-facility':             ['z2',  None,   None,   None],
            'zsyslog-server':               ['z2',  None,   None,   None],
            'zsyslog-access':               ['z2',  None,   None,   None],
            'zsyslog-access-facility':      ['z2',  None,   None,   None],
            'zsyslog-access-server':        ['z2',  None,   None,   None],
            'z2-log-file':                  ['z2',  '-l',   None,   None],
            'z2-pid-file':                  ['z2',  '--pid',None,   None]}

    def update (self, key, value):
        if self.__table.has_key(key):
            self.__table[key][self.__headers.index('value')] = value

    def get (self, key):
        if self.__table.has_key(key):
            return self.__table[key]
        return []

    def value (self, key):
        return self.__table[key][self.__headers.index('value')]

    def items (self):
        keys = self.__table.keys()
        keys.sort()
        return [ (key, self.__table[key][3]) for key in keys ]

    def keys (self):
        result = self.__table.keys()
        result.sort()
        return result

    def values (self):
        return [ item[3] for item in self.__table.values() ]

    def options (self, name):
        options = []
        for command, flag, _type, value in self.__table.values():
            if (command == name) and (value is not None) and (flag is not None):
                lower = value.lower()
                if   (_type == 'bool') and lower in self.__true:
                    options.extend([flag, '1'])
                elif (_type == 'bool') and lower in self.__false:
                    options.extend([flag, '0'])
                elif (_type == 'flag') and lower in self.__true:
                    options.append(flag)
                elif (_type == 'flag') and lower in self.__false:
                    pass
                elif (_type == 'multi'):
                    for item in value.split():
                        options.extend([flag, item])
                elif (_type == 'single'):
                    options.extend([flag, "'%s'" % value])
                else:
                    options.extend([flag, value])
        return options

    def env(self, name):
        env = []
        for key, (command, flag, _type, value) in self.__table.items():
            if (command == name) and (value is not None) and (flag is None):
                env.append('%s=%s' % (key.upper().replace('-','_'), value))
        return env


class Event(Exception):
    def notify(self):
        if not self.__dict__.has_key('notified'):
            sys.stdout.write(str(self))
            sys.stdout.flush()
            self.notified = 1

class ZopeInstanceEvent(Event):
    def __init__ (self, name, *args):
        event = {
            'kill':     _('could not send signal(%s): %s'),
            'running':  _('already running with pid(%s)'),
            'pidfile':  _('could not get PID from file(%s): %s'),
            'config':   _('unknown configuration key(%s)'),
            'zpasswd':  _('could not create access file(%s)'),
            'access':   _('access file(%s) is not a regular file'),
            'run':      _('system(%s) exited with status(%s)')
        }
        self.args = (name, event[name], args)

    def __str__ (self):
        return self.args[1] % self.args[2]

class ZopeInstance:
    """
    ZopeInstance(conf = {})

    Initialize a ZopeInstance by applying the values passed in 'conf'.
    Valid keys follow are those described in ZopeInstanceConf.
    """
    def __init__(self, conf = {}):
        self.__conf = ZopeInstanceConf()
        for option, value in conf.items():
            self.__conf.update(option, value)
        self.name = self.__conf.value('name')
        self.__command = {
            'pwd':  [ '/usr/sbin/zope-zpasswd' ],
            'z2':   [ '/usr/sbin/zope-z2' ]
        }
        self.__command['pwd'].extend(self.__conf.options('pwd'))
        self.__command['pwd'].append(self.__conf.value('access-file'))
        self.__command['z2'].extend(self.__conf.options('z2'))
        self.__command['z2'].extend(self.__conf.env('z2'))

    def __run(self, argv):
        command = ' '.join(argv)
        #print command
        status = os.system(command)
        if status:
            raise ZopeInstanceEvent('run', command[30:] + '...', status)

    def __install(self, mode, user, group, path):
        parent, None = os.path.split(path)
        if not os.path.exists(parent):
            self.__install(00755, 0, 0, parent)
        if not os.path.exists(path):
            #print 'mkdir %s' % path
            os.mkdir(path)
        #print 'chmod %s %s' % (mode, path)
        os.chmod(path, mode)
        #print 'chown %s.%s %s' %(user, group, path)
        os.chown(path, user, group)

    def __remove(self, path):
        if os.path.isdir(path):
            for entry in os.listdir(path):
                self.__remove(os.path.join(path, entry))
            #print 'rmdir %s' % path
            os.rmdir(path)
        elif os.path.exists(path):
            #print 'rm %s' % path
            os.remove(path)

    def __pid(self):
        pidfile = self.__conf.value('z2-pid-file')
        try:
            return int(open(pidfile,'r').read())
        except Exception, e:
            raise ZopeInstanceEvent('pidfile', pidfile, str(e))

    def __kill(self, signal):
        pid = self.__pid()
        try:
            #print 'kill %s %s' % (signal, pid)
            os.kill(pid, signal)
        except Exception, e:
            raise ZopeInstanceEvent('kill', signal, str(e))
        return pid

    def __start(self):
        log = self.__conf.value('z2-log-file')
        detailed_log = self.__conf.value('z2-detailed-log-file')
        pid = self.__conf.value('z2-pid-file')
        for filename in (log, detailed_log, pid):
            self.__install(00755, 0, 0, os.path.split(filename)[0])
        self.__run(self.__command['z2'])
        sleep(1)

    def __str__(self):
        try:
            pid = self.__kill(0)
        except:
            pid = None
        result = [ '%s(%s):' % (self.name, pid )] + [
            '\t%s: %s' % (key, value) for key, value in self.__conf.items()
        ]
        return '\n'.join(result)
    
    def isrunning(self):
        """
        isrunning() -> value

        If the instance is running, 'value' will be 1, else it will be 0.
        """
        if os.path.exists(self.__conf.value('z2-pid-file')):
            try:
                self.__kill(0)
            except ZopeInstanceEvent, e:
                return 0
            else:
                return 1
        else:
            return 0

    def create(self):
        """
        create() -> None

        Create all the path and files needed to run the instance.
        """
        instance_home = self.__conf.value('instance-home')
        data_fs_in = self.__conf.value('data-fs-in')
        data_fs = os.path.join(instance_home, 'var', 'Data.fs')
        group = getgrnam(self.__conf.value('group'))[2]
        user = getpwnam(self.__conf.value('user'))[2]
        self.__install(00755, 0, 0,     os.path.join(instance_home, 'Extensions'))
        self.__install(00755, 0, 0,     os.path.join(instance_home, 'import'))
        self.__install(01755, 0, group, os.path.join(instance_home, 'var'))
        #print 'cp %s %s' % (data_fs_in, data_fs)
        open(data_fs,'w').write(open(data_fs_in,'r').read())
        #print 'chown %s.%s %s' % (user, group, data_fs_in)
        os.chown(data_fs, user, group)
        #print 'chmod 00644 %s' % data_fs
        os.chmod(data_fs, 00644)

    def access(self):
        self.__run(self.__command['pwd'])

    def pcgi(self):
        pcgi_resources = [
            '#!%s' % self.__conf.value('pcgi-wrapper'),
            '# Automatically generated by zopectl(8) command.',
            '# Do _NOT_ edit!'
        ] + self.__conf.env('pcgi')
        pcgi_resource_file = self.__conf.value('pcgi-resource-file')
        #print 'cat > %s' % pcgi_resource_file
        open(pcgi_resource_file,'w').write('\n'.join(pcgi_resources))
        #print 'chmod 00644 %s' % pcgi_resource_file
        os.chmod(pcgi_resource_file, 00644)

    def delete(self):
        """
        delete() -> None

        Stops the instance and remove all the path and files used to run it.
        """
        try:
            self.stop();
        except:
            pass
        self.__remove(self.__conf.value('instance-home'))

    def start(self):
        """
        start() -> None

        Starts the instance using the z2.py command with a proper argv.
        If instance is already running, a proper ZopeInstanceEvent is raised.
        """
        if not self.isrunning():
            self.__start()
        else:
            raise ZopeInstanceEvent('running', self.__pid())

    def stop(self):
        """
        stop() -> None

        Stops the instance by sending SIGTERM to its PID.
        """
        self.__kill(SIGTERM)

    def restart(self):
        """
        restart() -> None

        If instance is running, restarts it by sending SIGHUP to its PID; in
        any other case starts the instance.
        """
        if self.isrunning():
            self.__kill(SIGHUP)
        else:
            self.__start()

    def logrotate(self):
        """
        logrotate() -> None

        If instance is running, force it to reopen its log files by sending
        SIGUSR2 to its PID.
        """
        self.__kill(SIGUSR2)

    def status(self):
        print self

class ZopeControlEvent(Event):
    def __init__ (self, name, *args):
        event = {
            'failed':       _(' failed: %s\n'),
            'timeout':      _(' timed out (failure?): %s\n'),
            'arrived':      _(' %s(%s),'),
            'done':         _(' done.\n'),
            'start':        _('Starting Zope instance %s...\n'),
            'stop':         _('Stopping Zope instance %s...\n'),
            'restart':      _('Restarting Zope instance %s...\n'),
            'logrotate':    _('Forcing logs reopening for Zope instance %s...'),
            'status':       _('Dumping Zope Instances: %s\n'),
            'create':       _('Creating home for Zope instance %s...'),
            'access':       _('Creating emergency user for Zope instance %s...\n'),
            'pcgi':         _('Creating PCGI resource file for Zope instance %s...'),
            'delete':       _('Deleting Zope Instance %s...'),
            'wait':         _(' Waiting instances in late:'),
        }
        self.args = (name, event[name], args)

    def __str__ (self):
        return self.args[1] % self.args[2]

class ZopeControlFailure(ZopeControlEvent):
    def __init__ (self, action, instance, error):
        self.args = ('failure', action, instance, error)

    def __str__ (self):
        return _('%s: %s: %s') % (self.args[1:])

class ZopeControlDisaster(ZopeControlEvent):
    def __init__(self, events):
        self.args = ('disaster', events)

    def __str__(self):
        events = [ str(event) for event in self.args[1]]
        events.sort()
        return '\n'.join(events)
        

class ZopeControl:
    """
        ZopeControl(selfrc, paths)

        Initialize the ZopeControl instance by loading the ZopeControl RC
        file, and all the configuration paths. It then builds a list of
        ZopeInstance instances based on the value found in the various
        paths. Wrong paths are silently ignored.
        selfrc should be a valid configuration files with 2 configuration
        sections:
            instances:
                The default value to pass to each instance (usually, those
                in [REQUIRED] section in the ZopeInstanceConf documentation.
            zopectlrc:
                This is the section dedicated to configuration of
                ZopeControl instance (currently no option has been defined).
    """
    def __init__ (self, selfrc, paths):
        conf = ConfigParser()
        conf.read(selfrc)
        self.__conf = {'instances': {}, 'zopectlrc': {}}
        if conf.has_section('instances'):
            for option in conf.options('instances'):
                self.__conf['instances'][option] = conf.get('instances', option, 1)
        self.paths = paths
        self._load()

    def __getopt (self, argv):
        self.name = os.path.splitext(os.path.basename(argv[0]))[0]
        self.actions = []
        flags = (
            ('-s', '--start'),
            ('-S', '--stop'),
            ('-r', '--restart'),
            ('-f', '--force-reload'),
            ('-c', '--create'),
            ('-a', '--access'),
            ('-p', '--pcgi'),
            ('-d', '--delete'),
            ('-l', '--logrotate')
        )
        short = ':'.join([ flag[0][1] for flag in flags ]) + ':'
        _long = [ flag[1][2:] + '=' for flag in flags]
        (options, self.args) = getopt(argv[1:], short, _long)
        for option, instances in options:
            for pair in flags:
                if option in pair:
                    action = pair[1][2:]
                    self.actions.append([action, instances.split(',')])
        any = [ instance.name for instance in self.instances ]
        for index in range(len(self.actions)):
            if self.actions[index][1] == ['*']:
                self.actions[index][1] = any
        if self.args:
            action = self.args[0]
            if action in [ pair[1][2:] for pair in flags ]:
                instances = self.args[1:]
                if instances:
                    self.actions.append([action, instances])
                else:
                    self.actions.append([action, any])

    def _load (self, names = None):
        """
            _load(names = None) -> None

            Load the configuration options for each instance name in 'names'.
            If names is None, all available instances are loaded. Unknown
            values in 'names' are ignored.
        """
        self.instances = []
        defaults = self.__conf['instances'].copy()
        for path in self.paths:
            defaults['basename'] = os.path.splitext(os.path.split(path)[1])[0]
            conf = ConfigParser(defaults)
            conf.read(path)
            for section in conf.sections():
                var = {'section': section}
                if (names is None) or (conf.get(section, 'name') in names):
                    options = {}
                    for name in conf.defaults().keys():
                        options[name] = conf.get('DEFAULT', name, 0, var)
                    for name in conf.options(section):
                        options[name] = conf.get(section, name, 0, var)
                    self.instances.append(ZopeInstance(options))

    def _apply (self, action, instances, timeout = 0):
        """
            _apply(action, instances, timeout = 0) -> None

            Apply action 'action' to each instance in 'instances'. If
            'timeout' is greater then 0, it waits for each action to be
            completed.
        """
        late = []
        loaded = [ instance.name for instance in self.instances ]
        for instance in instances:
            if instance in loaded:
                late.append(self.instances[loaded.index(instance)])
        failures = []
        for instance in late[:]:
            ZopeControlEvent(action, instance.name).notify()
            try:
                eval('instance.%s()' % action)
            except Exception, e:
                ZopeControlEvent('failed', e).notify()
                late.remove(instance)
                failures.append(ZopeControlFailure(action, instance.name, e))
            else:
                if not timeout:
                    ZopeControlEvent('done').notify()
        if timeout > 0:
            ZopeControlEvent('wait').notify()
            for time in range(timeout):
                sleep(1)
                for instance in late[:]:
                    running = instance.isrunning()
                    if (action == 'stop' and not running) or running:
                        ZopeControlEvent('arrived', instance.name, time).notify()
                        late.remove(instance)
                if not late:
                    ZopeControlEvent('done').notify()
                    break
            for instance in late:
                failures.append(ZopeControlFailure('timeout', instance.name, timeout))
        if len(failures) > 1:
            # ... Quanto mi piace queststo "Disaster"!... :^P
            # Non vedevo l'ora di impiegare una parola come questa :^D
            raise ZopeControlDisaster(failures)
        elif failures:
            raise failures[0]

    def run (self, argv):
        """
    usage:
            %s options [ action [ [ instance_name ] ... ] ]
            %s [ options ]  action [ [ instance_name ] ... ]
    options:
        Each option need an argument. Each argument must be a comma separated
        list of instance names. Unknown instance names are ignored. The special
        '*' instance name means 'any known instance'.
        For each instance_name resulting from option parsing:
            o -s, --start           Apply action 'start'
            o -S, --stop            Apply action 'stop'
            o -r, --restart         Apply action 'restart'
            o -f, --force-reload    Apply action 'force-reload'
            o -c, --create          Apply action 'create'
            o -a, --access          Apply action 'access'
            o -p, --pcgi            Apply action 'pcgi'
            o -d, --delete          Apply action 'delete'
            o -l, --logrotate       Apply action 'logrotate'
    action:
        For each instance_name (all, if no one is specified):
            o start         Start instance_name
            o stop          Stops instance_name
            o restart       Restarts instance_name
            o force-reload  Force instance_name configuration reloading
            o logrotate     Force instance_name to reopen logs
            o create        Initialize instance_name files and directories
            o access        Regenerate emergency user for instance_name
            o pcgi          Regenerate PCGI resource file for instance_name
            o delete        Stops and remove instance_name files and directories
        """
        steps = {
            'start':        ('self._apply(action, instances, 30)',),
            'stop':         ('self._apply(action, instances, 5)',),
            'restart':      ('self._apply(action, instances, 30)',),
            'logrotate':    ('self._apply(action, instances)',),
            'status':       ('self._apply(action, instances)',),
            'access':       ('self._apply(action, instances)',),
            'delete':       ('self._apply(action, instances)',),
            'status':       ('self._apply(action, instances)',),
            'pcgi':         ('self._apply(action, instances)',),
            'create':       (
                "self._apply('create', instances)",
                "self._apply('access', instances)",
                "self._apply('pcgi', instances)"
            ),
            'force-reload': (
                "self._apply('stop', instances, 5)",
                "self._load(instances)",
                "self._apply('start', instances, 30)"
            )
        }
        self.__getopt(argv)
        for action, instances in self.actions:
            failures = []
            for step in steps[action]:
                try:
                    eval(step)
                except ZopeControlEvent, e:
                    if e.args[0] == 'disaster':
                        failures.extend(e.args[1])
                    else:
                        failures.append(e)
            if len(failures) > 1:
                raise ZopeControlDisaster(failures)
            elif failures:
                raise failures[0]
        return 0

    def status(self, event):
        """
        status(event) -> exit_status

        Returns the exit status value associated with the event, prints the
        event message and run.__doc__ if needed.
        """
        event.notify()
        error = {
            'usage':    1,
            'failure':  2,
            'disaster': 3,
            'other':    4
        }
        if isinstance(event, ZopeControlEvent):
            status = error[event.args[0]]
            if event.args[0] == 'usage':
                print _(self.run.__doc__) % self.name
        else:
            status = error['other']
        return status
